#!/usr/bin/env python3.8
# -*- coding: utf-8 -*-

"""This script contains Logger instance to print data with timestamp on console or log to file
"""

# set timezone for data ingestion
from fp_th_di.date_utils import *

# Logger tool to print data with timestamp
class Logger:
  def appendLog(self, filename:str, *txt):
    with open(filename, 'a') as f:
      txt = [str(t) for t in txt]
      f.write(','.join(txt) + '\n')

  def log_info_with_timestamp(self, filename:str, *txt):
    with open(filename, 'a') as f:
      txt = [str(t) for t in txt]
      f.write('[INFO] ' + now().strftime('%Y-%m-%d %X') + ' - ' + ' '.join(txt) + '\n')

  def log_error_with_timestamp(self, filename:str, *txt):
    with open(filename, 'a') as f:
      txt = [str(t) for t in txt]
      f.write('[ERROR] ' + now().strftime('%Y-%m-%d %X') + ' - ' + ' '.join(txt) + '\n')

  def writeToFile(self, filename:str, txt:str):
    with open(filename, 'w') as f:
      f.write(txt)

  def info(self, *txt):
    txt = [str(t) for t in txt]
    print('[INFO]', now().strftime('%Y-%m-%d %X'), '-', ' '.join(txt))

  def error(self, *txt):
    txt = [str(t) for t in txt]
    print('[ERROR]', now().strftime('%Y-%m-%d %X'), '-', ' '.join(txt))

  def exception(self, *txt):
    txt = [str(t) for t in txt]
    print('[ERROR]', now().strftime('%Y-%m-%d %X'), '-', ' '.join(txt))

  def logInfo(self, filename:str, *txt):
    txt = [str(t) for t in txt]
    logStr = '[INFO]' + now().strftime('%Y-%m-%d %X') + ' - ' + ' '.join(txt)
    self.appendLog(filename, logStr)

  def logError(self, filename:str, *txt):
    txt = [str(t) for t in txt]
    logStr = '[ERROR]' + now().strftime('%Y-%m-%d %X') + ' - ' + ' '.join(txt)
    self.appendLog(filename, logStr)

logger = Logger()



from googleapiclient.discovery import Resource, build
from apiclient.http import MediaFileUpload
from fp_th_di.logger import logger
from google.oauth2 import service_account
from googleapiclient.http import MediaIoBaseDownload
import io
import pandas as pd

def get_google_drive_credential(
    serviceAccountSecretCredentialsLocation:str,
    scopes:list
  ):
  """Get google drive credential or create new one if not found
  """
  credentials = service_account.Credentials.from_service_account_file(
        serviceAccountSecretCredentialsLocation, scopes=scopes)
  return credentials

def create_google_drive_service(
    serviceName:str,
    serviceAccountSecretCredentialsLocation:str,
    apiVersion:str,
    scopes:list
  ):
  return build(
    serviceName,
    apiVersion,
    credentials=get_google_drive_credential(
      serviceAccountSecretCredentialsLocation=serviceAccountSecretCredentialsLocation,
      scopes=scopes
    )
  )

def generate_google_drive_folder_metadata(folderName:str, parentFolderId:str) -> dict:
  """Generates metadata dict for folder type

  Args:
    folderName (str): folder name to create
    parentFolderId (str): google id of parent folder to create new folder in

  Returns:
    dict: the folder metadata
  """
  return {
    'name': folderName,
    'mimeType': 'application/vnd.google-apps.folder',
    'parents': [parentFolderId]
  }

def create_new_google_drive_folder(driveService:Resource, folderName:str, parentFolderId:str) -> str:
  """Creates new google drive folder

  Args:
    driveService (Resource): google drive service
    folderName (str): folder name
    parentFolderId (str): google id of parent folder to create new folder in

  Returns:
    str: new folder id
  """
  folder = driveService.files().create(fields='id',
    body=generate_google_drive_folder_metadata(
      folderName=folderName,
      parentFolderId=parentFolderId
    )).execute()
  logger.info('Successfully created new Google Drive folder:', folder.get('id'))
  return folder.get('id')

def search_for_google_drive_folder(driveService:Resource, folderName:str) -> str:
  """Search for targeted folder by name

  Args:
    driveService (Resource): google drive service
    folderName (str): name of folder to search for

  Returns:
    str: [description]
  """
  logger.info('Searching for Google Drive Folder:', folderName)
  pageToken = None
  while True:
    response = driveService.files().list(
      q="mimeType = 'application/vnd.google-apps.folder'",
      spaces='drive',
      fields='nextPageToken, files(id, name)',
      pageToken=pageToken
      ).execute()
    for file in response.get('files', []):
      if file.get('name') == folderName:
        logger.info('Found folder: %s (%s)' % (file.get('name'), file.get('id')))
        return file.get('id')
    pageToken = response.get('nextPageToken', None)
    if pageToken is None:
      break
  logger.info('Could not found targeted folder in your Drive')

def list_folders_in_drive_folder(driveService, parentFolderId: str) -> list:
  """_summary_
  List out GoogleDrive folders exist in the parent folder with id given as input parameter

  Args:
      driveService: Google Drive Service object
      parentFolderId (str): Parent Google Drive Folder Id

  Returns:
      list: list of dictionary with information of folders found in the Shared Drive
  """
  logger.info("Getting folders from:", parentFolderId)
  results = driveService.files().list(
      q=f"mimeType = 'application/vnd.google-apps.folder' and trashed = false and parents in '"+ parentFolderId +"'",
      spaces="drive",
      fields="nextPageToken, files(id, name, createdTime)",
      orderBy="createdTime"
  ).execute()
  items = results.get('files', [])

  if not items:
      return ""
  else:
      return items

def generate_file_metadata(filename:str, parentFolderId:str, mimeType:str) -> dict:
  """Generates metadata dict for file type

  Args:
    filename (str): file to create
    parentFolderId (str): google id of parent folder to create the file in
    mimeType (str): file mime type. For example:
        image/png or image/jpeg for images
        application/vnd.openxmlformats-officedocument.spreadsheetml.sheet for xlsx file
        application/vnd.ms-excel for xls file
        text/csv for csv file

  Returns:
    dict: the folder metadata
  """
  return {
    'name': filename,
    'mimeType': mimeType,
    'parents': [parentFolderId]
  }

def search_files_from_google_drive_folder_by_filename(driveService:Resource, folderId:str, fileMimeType:str=None, targetedFilename:str=None) -> list:
  """Searching for files by mimetype and filename in specific GoogleDrive folder

  Args:
      driveService (Resource): google drive service
      folderId (str): Google drive folder Id
      fileMimeType (str, optional): file mime type. For example:
        image/png or image/jpeg for images
        application/vnd.openxmlformats-officedocument.spreadsheetml.sheet for xlsx file
        application/vnd.ms-excel for xls file
        text/csv for csv file
        Defaults to None.
      targetedFilename (str, optional): Filename to search. Defaults to None. If filename's none, get all files in the folder

  Returns:
      list: list of files dictionary. An item in the list consist of id of the file and filename
  """
  logger.info("Listing files in folder:", folderId)

  mimeType = f"mimeType = '{fileMimeType}' and " if fileMimeType is not None else ""
  filename = f"name = '{targetedFilename}' and " if targetedFilename is not None else ""

  pageToken = None
  results = driveService.files().list(
    q=f"{mimeType} {filename} trashed = false and parents in '{folderId}'",
    spaces='drive',
    fields='nextPageToken, files(id, name)',
    pageToken=pageToken,
    orderBy="createdTime desc",
  ).execute()

  return results.get('files', [])

def upload_file_to_google(driveService:Resource, localFileLocation:str, filename:str, parentFolderId:str, mimeType:str) -> str:
  """Upload a file from local to google drive folder

  Args:
    driveService (Resource): google drive service
    localFileLocation (str): path to file location
    filename (str): filename
    parentFolderId (str):  google id of parent folder to create the file in
    mimeType (str): file mime type. For example:
      image/png or image/jpeg for images
      application/vnd.openxmlformats-officedocument.spreadsheetml.sheet for xlsx file
      application/vnd.ms-excel for xls file
      text/csv for csv file

  Returns:
    str: google id of the file uploaded to Google Drive
  """
  media = MediaFileUpload(f'{localFileLocation}/{filename}', mimetype=mimeType)

  # Upload the file, use supportsAllDrives=True to enable uploading if targeted parent folder is shared drives.
  file = driveService.files().create(
    body=generate_file_metadata(
      filename=filename,
      parentFolderId=parentFolderId,
      mimeType=mimeType
    ),
    media_body=media,
    supportsAllDrives=True
  ).execute()
  logger.info("Successfully created file '%s' id '%s'." % (file.get('name'), file.get('id')))
  return file.get('id')

def get_folder_in_shared_drive(driveService, sharedDriveId: str) -> list:
    """_summary_

    Args:
        driveService: Google Drive Service object
        sharedDriveId (str): driveService: Google Drive Service object

    Returns:
        list: list of dictionary with information of folders found in the Shared Drive
    """
    results = driveService.files().list(
        q=f"mimeType = 'application/vnd.google-apps.folder' and trashed = false",
        driveId=sharedDriveId,
        includeItemsFromAllDrives=True,
        corpora='drive',
        fields="nextPageToken, files(id, name, createdTime)",
        orderBy="createdTime",
        supportsAllDrives=True
    ).execute()
    items = results.get('files', [])

    if not items:
        return ""
    else:
        return items


def get_folder_id_from_shared_drive_by_folder_name(driveService,
    folderName:str, sharedDriveId: str) -> str:
    """Get Google Drive folder if from Shared Drive by searching from folder name

    Args:
        driveService: Google Drive Service object
        folderName (str): Targeted folder name
        sharedDriveId (str): driveService: Google Drive Service object

    Returns:
        str: id of the Google Drive folder if found in the Shared Drive; otherwise, None.
    """
    folders = get_folder_in_shared_drive(driveService, sharedDriveId)
    if not folders:
      return None
    for each in folders:
      if each['name'] == folderName:
        return each['id']
    else:
      return None

def get_files_from_shared_drive_by_folder_id(driveService,
    folderId: str, sharedDriveId: str, fileMimeType: str = None,
    targetedFilename: str = None) -> list:
    """Get files from Shared Drive by Google Drive folder id

    Args:
        driveService: Google Drive Service object
        folderId (str): id of gogole drive folder
        sharedDriveId (str): id of shared drive folder
        fileMimeType (str, optional): targeted files' type. Defaults to None.
        targetedFilename (str, optional): targeted files' names. Defaults to None.

    Returns:
        list: list of dictionary with information of targeted files found in the Google Drive folder.
    """
    mimeType = f"mimeType = '{fileMimeType}' and" if fileMimeType is not None else ""
    filename = f"name = '{targetedFilename}' and " if targetedFilename is not None else ""

    results = driveService.files().list(
        q=f"{mimeType} {filename} parents in '{folderId}' and trashed = false",
        driveId=sharedDriveId,
        includeItemsFromAllDrives=True,
        corpora='drive',
        fields="nextPageToken, files(id, name, createdTime)",
        orderBy="createdTime desc",
        supportsAllDrives=True
    ).execute()
    return results.get('files', [])

def delete_file_in_drive_by_file_id(driveService, fileId:str, folderId:str):
  """Remove file with Id from google drive folder Id

  Args:
      driveService (Resource): google drive service
      fileId (str): file id
      folderId (str): google drive folder id
  """
  driveService.files().update(fileId=fileId, removeParents=folderId).execute()

def download_file_from_drive_by_file_id(driveService, fileId:str, localFilePath:str):
    """Download targeted file from Google Drive by file id

    Args:
        driveService: Google Drive Service
        fileId (str): targeted file id
        localFilePath (str): path to save file locally
    """
    request = driveService.files().get_media(fileId=fileId)
    fh = io.BytesIO()
    downloader = MediaIoBaseDownload(fh, request)
    done = False
    while done is False:
        status, done = downloader.next_chunk()
        print("Download %d%%." % int(status.progress() * 100))
    # Write the stuff
    with open(localFilePath, "wb") as f:
        f.write(fh.getbuffer())

def read_dataframe_from_google_sheet(sheetService, fileId:str, range:str) -> pd.DataFrame:
    """Return dataframe from specified range in targeted Google Sheet

    Args:
        sheetService (object): Google service with credential
        fileId (str): GoogleSheet File ID
        range (str): Range of information to extract into dataframe

    Returns:
        pd.DataFrame: dataframe with the information from the range specified in the input
    """
    # Get dataframe from file
    sheetObject = sheetService.spreadsheets().values().get(
      spreadsheetId=fileId,
      range=range
    ).execute()
    valuesFromRanges = sheetObject.get("values", [])
    return pd.DataFrame(valuesFromRanges[1:], columns=valuesFromRanges[0])

def write_dataframe_in_google_sheet(
    sheetService, fileId:str,
    range:str, dataframe:pd.DataFrame
  ):
  """Update gsheet range with values from dataframe

  Args:
      sheetService (_type_): Google Drive Service
      fileId (str): gsheet file id
      range (str): range to update
      dataframe (pd.DataFrame): dataframe to update in the range
  """
  try:
    # clearing out current values
    logger.info("Clearing out previous data")
    sheetService.spreadsheets( ).values( ).clear(
      spreadsheetId=fileId,
      range=range,
      body={}).execute( )

    logger.info("Writing new data")
    values = [list(dataframe.columns)]
    if len(dataframe) > 0:
      values.extend(dataframe.values.tolist())

    body = {
        'values': values
    }
    result = sheetService.spreadsheets().values().update(
        spreadsheetId=fileId,
        range=range,
        valueInputOption="USER_ENTERED",
        body=body
    ).execute()
    logger.info(f"{result.get('updatedCells')} cells updated.")
    return result

  except Exception as error:
    logger.error(f"An error occurred: {error}")
    return error

def create_google_sheet_service(serviceAccountSecretCredentialsLocation:str):
  return create_google_drive_service(
    serviceName='sheets',
    serviceAccountSecretCredentialsLocation=serviceAccountSecretCredentialsLocation,
    apiVersion='v4',
    scopes=['https://www.googleapis.com/auth/spreadsheets']
  )

def create_google_drive_service_sh(serviceAccountSecretCredentialsLocation:str):
  return create_google_drive_service(
    serviceName="drive",
    serviceAccountSecretCredentialsLocation=serviceAccountSecretCredentialsLocation,
    apiVersion="v3",
    scopes=['https://www.googleapis.com/auth/drive']
  )



import imaplib
import email
import email.utils 
import os
from email.header import decode_header
from email.mime.multipart import MIMEMultipart 
from email.mime.text import MIMEText 
from email.mime.base import MIMEBase 
from email.mime.image import MIMEImage
from email.mime.audio import MIMEAudio
from fp_th_di.logger import logger
from typing import Callable, Union, List
from fp_th_di.mail_box.email_object import Email
from fp_th_di.mail_box.email_attachment import EmailAttachment
import fp_th_di.mail_sender  as ms

class MailBox:
  def __init__(self, username:str, password:str, directory:str) -> None:
    """Initialize mailbox

    Args:
        username (str): email address 
        password (str): email password
        directory (str): download directory for mail attachments
    """    
    self.username = username
    self.password = password
    self.directory = directory
    self.imap = None

  def login(self) -> None:
    """Login to Gmail"""
    try:
      self.imap = imaplib.IMAP4_SSL("imap.gmail.com")
      self.imap.login(self.username, self.password)
    except Exception as e:
      logger.exception(e)
      self.imap = None

  def logout(self) -> None:
    """Logout from Gmail"""    
    try:
      self.imap.close()
      self.imap.logout()
      self.imap = None
    except Exception as e:
      logger.exception(e)

  def is_logged_in(self) -> bool:
    """Check whether the connection to IMAP server's ready

    Returns:
        bool: True if mailbox's logged in; otherwise, False
    """    
    return not (self.imap is None)

  def get_emails_from_inbox(self, searchCriteria:str = "ALL", fetchLimit: Union[int, None] = 1, 
      boxName:str='INBOX', filteringFunction:Callable[[Email], bool] = None) -> List[Email]:
    """Get latest emails that meet the filteringFunction from targeted mailbox 

    Args:
        searchCriteria (str, optional): IMAP Search criteria to apply. Defaults to "ALL".
        fetchLimit (Union[int, None]|, optional): limit number of emails to fetch. Defaults to 1.
        boxName (str, optional): Targeted mailbox or tag to search from. Defaults to 'INBOX'.
        filteringFunction (Callable[[Email], bool], optional): filter function to applied when search. Defaults to None.

    Raises:
        Exception: raises 'Connection to mailbox is not ready' if mailbox was not logged in to IMAP 

    Returns:
        List[Email]: list of emails that met the search criteria found in targeted mailbox
    """      
    if not self.is_logged_in():
      raise Exception("Connection to mailbox is not ready")

    # search for emails that meet the searchCriteria
    logger.info('LOG - MailBox: Searching for ' + searchCriteria + ' in ' + boxName)
    typ, selectdata = self.imap.select(boxName, readonly=True)
    if typ == 'NO':
      return []
    
    status, messages = self.imap.search(None, searchCriteria)
    messages = messages[0].split()
    messages.reverse()

    emails = []
    for i in messages:
      # stop fetching emails if reach fetch limit
      if (fetchLimit is not None and len(emails) == fetchLimit):
        break

      # fetch each email message by ID
      res, msg = self.imap.fetch(i, "(RFC822)")
      for response in msg:
        incomingMail = None
        if isinstance(response, tuple):
          # parse a bytes email into a message object
          msg = email.message_from_bytes(response[1])
          # decode the email subject
          subject = decode_header(msg["Subject"])[0][0]
          if isinstance(subject, bytes):
            # if it's a bytes, decode to str
            subject = subject.decode()
          
          # email sender and timestamp
          from_ = msg.get("From")
          receivedDate=email.utils.parsedate(msg.get('date'))
          
          # initialize email object
          incomingMail = Email(from_, receivedDate, subject)
          incomingMail.emailAttachments = []

          # if the email message is multipart
          if msg.is_multipart():
            # iterate over email parts
            for part in msg.walk():
              # extract content type of email
              content_type = part.get_content_type()
              content_disposition = str(part.get("Content-Disposition"))
              body = ""
              try:
                # get the email body
                body = part.get_payload(decode=True).decode()
              except:
                pass
              
              if content_type == "text/plain" and "attachment" not in content_disposition:
                incomingMail.email_content = body
              elif "attachment" in content_disposition:
                # download attachment
                filename = part.get_filename()
                filepath = f'{self.directory}/{filename}'
                dummy, fileExtension = os.path.splitext(filename)
                if filename:
                  with open(filepath, "wb") as attachmentPayload:
                    attachmentPayload.write(part.get_payload(decode=True))
                  
                  if (fileExtension.lower() in ('.zip', '.csv', '.txt', '.xlsx')):
                    incomingAttachment = EmailAttachment(filepath, filename, fileExtension)
                    incomingMail.add_attachment(incomingAttachment)
          else:
            content_type = msg.get_content_type()
            body = msg.get_payload(decode=True).decode()
            if content_type == "text/plain":
              incomingMail.emailContent = body
          
          if incomingMail:
            # discard the email of not pass filtering funtion
            if filteringFunction is not None and not filteringFunction(incomingMail):
              break

            emails.append(incomingMail)
            break
          
    return emails

  def send_mail(self, receiver:Union[str,list], subject:str, 
    content:str, isHTML:bool = False, attachments:list = []) -> None:
    """Send outgoing email

    Args:
        receiver (Union[str,list]): receivers
        subject (str): email subject
        content (str): email body
        isHTML (boolean): is the email content in HTML format
        attachments (list, optional): list of path_to_attachments. Defaults to [].
    """    
    logger.info('LOG - mailbox: sendmail - ' + subject + ' - To:' + str(receiver) + ' Attachments: ' + str(attachments))
    try:
      ms.send_mail(
        email_from=self.username,
        email_password=self.password,
        email_to=receiver, 
        subject=subject,  
        content=content,    
        is_html=isHTML,        
        attachments=attachments
      )
    except Exception as e:
      logger.exception(e)



class EmailAttachment:
  def __init__(self, path:str, filename:str, filetype:str):
    self.filepath = path
    self.filename = filename
    self.filetype = filetype

  def __eq__(self, obj) -> bool:
    return self.filepath == obj.filepath and self.filetype == obj.filetype \
      and self.filename == obj.filename
  
  def __str__(self) -> str:
    return str(vars(self))

  def __getitem__(self, key:str):
    return self.__getattribute__(key)




from typing import List
import pysftp
import re
import os
from fp_th_di.logger import logger

class SFTPManager:
  def __init__(self, host: str, username: str, password: str=None, privateKey: str=None,
      privateKeyPass:str=None, port:int=22):
    self.host = host
    self.username = username
    self.password = password
    self.privateKey = privateKey
    self.privateKeyPass = privateKeyPass
    self.port = port

    self.cnopts = pysftp.CnOpts()
    self.cnopts.hostkeys = None

  def create_connection(self):
    try:
      if self.privateKey is not None:
        self.sftp = pysftp.Connection(host=self.host, username=self.username,
            private_key=self.privateKey, private_key_pass=self.privateKeyPass,
            port=self.port, cnopts=self.cnopts)
      elif self.password is not None:
        self.sftp = pysftp.Connection(host=self.host, username=self.username,
            password=self.password, cnopts=self.cnopts, port=self.port)
      else:
        raise Exception('Please provide sftp password or private key')
    except Exception as e:
      logger.error(e)
      raise e

  def close_connection(self):
    try:
      self.sftp.close()
    except Exception as e:
      logger.error(e)
      raise e

  def list_files_in_path(self, remotePath:str = '.') -> list:
    """Get list of filenames in the given remote directory

    Args:
        remotePath (str): remote directory

    Raises:
        Exception: if failed to connect to SFTP server

    Returns:
        list: List of filenames found in remotePath
    """
    try:
      return list(self.sftp.listdir(remotePath))
    except Exception as e:
      logger.error(e)
      raise e

  def download_file(self, remoteFilename:str, localFilename:str):
    """Download file from remote directory to local directory

    Args:
        remoteFilename (str): path_to_remote_file_to_download_from
        localFilename (str): path_to_local_file_to_download_to

    Raises:
        Exception: if failed to connect to SFTP server
    """
    try:
      self.sftp.get(remoteFilename, localFilename)
    except Exception as e:
      logger.error(e)
      raise e

  def __downloadDir__(self, sftp:pysftp.Connection, filesInDir:List[str], localPath:str, filterOptions: dict = dict()) -> List[str]:
    """A helper private function to download all files given in the list of filesInDir

    Args:
        sftp (pysftp.Connection): SFTP connection
        filesInDir (List[str]): list of filenames to download
        localPath (str): local path to download to
        filterOptions (dict, optional): optional filtering condition to detemine whether a file should be downloaded or not. Defaults to dict().

    Returns:
        List[str]: list of path_to_downloaded_files
    """
    files = []
    for eachFile in filesInDir:
      filename, fileExtension = os.path.splitext(eachFile)
      # not download if does not pass filtering conditions
      if ('name_reg' in filterOptions) and \
        not bool(re.match(filterOptions['name_reg'], eachFile)):
        continue
      if ('file_format' in filterOptions) and \
        (not filterOptions['file_format'] in fileExtension):
        continue
      if ('name_like' in filterOptions) and \
        (not filterOptions['name_like'] in filename):
        continue
      # download file
      sftp.get(eachFile, localPath + eachFile)
      files.append(localPath + eachFile)
    return files

  def download_dir(self, remotePath:str, localPath:str, filterOptions: dict = dict()) -> List[str]:
    """Download all files in remote directory to a specific local directory

    Args:
        remotePath (str): remote directory to download from
        localPath (str): local directory to download to
        filterOptions (dict, optional): filtering option to determine whether a file should be downloaded or not. Defaults to dict().

    Raises:
        Exception: if failed to connec to SFTP server

    Returns:
        list: list of path_to_downloaded_files
    """
    try:
      self.__downloadDir__(self.sftp, self.list_files_in_path(remotePath), localPath, filterOptions)
    except Exception as e:
      logger.error(e)
      raise e

  def upload_file(self, localFilename:str, remoteFilename:str):
    """Upload local file to SFTP server

    Args:
        localFilename (str): path_to_local_file_to_upload
        remoteFilename (str): path_to_remote_file

    Raises:
        Exception: if failed to connect to SFTP server
    """
    try:
      self.sftp.put(localFilename, remoteFilename)
    except Exception as e:
      logger.error(e)
      raise e

  def delete_file(self, remoteFilename:str):
    """delete a specific file in SFTP server

    Args:
        remoteFilename (str): path_to_remote_file

    Raises:
        Exception: if fail to connect to SFTP server
    """
    try:
      self.sftp.remove(remoteFilename)
    except Exception as e:
      logger.error(e)
      raise e



import os, shutil, gzip
from os import listdir
from os.path import isfile, join
from typing import List
from fp_th_di.logger import logger
import re
  
def current_dir() -> str:
  """Get current directory

  Returns:
      str: path to current directory
  """  
  return os.getcwd()

def dir_exists(pathToDir:str) -> bool:
  """Check if directory exists

  Args:
      pathToDir (str): targeted path to directory to check

  Returns:
      bool: True if exists; otherwise, False
  """  
  return os.path.exists(pathToDir)

def clear_dir(pathToDir:str, options: dict=dict()) -> None:
  """Clear all files in targeted directory

  Args:
      pathToDir (str): targeted path to directory
      options (dict, optional): file format to deleted. Defaults to empty dict.
  """  
  for filename in os.listdir(pathToDir):
    filePath = os.path.join(pathToDir, filename)
    try:
      if os.path.isfile(filePath) or os.path.islink(filePath):
        if 'file_format' in options and \
          not options['file_format'].lower() in filePath.lower():
          continue
        os.unlink(filePath)
      elif os.path.isdir(filePath):
        shutil.rmtree(filePath)
    except Exception as e:
        logger.info('Failed to delete %s. Reason: %s' % (filePath, e))

def create_dir(pathToDir:str) -> None:
  """Create directory

  Args:
      pathToDir (str): path to directory to create
  """  
  try:
    os.makedirs(pathToDir)
  except OSError:
    logger.info("Creation of the directory %s failed" % pathToDir)
  else:
    logger.info("Successfully created the directory %s " % pathToDir)

def delete_dir(pathToDir:str) -> None:
  """Delete directory

  Args:
      pathToDir (str): targeted path to directory to delete
  """  
  try:
    os.rmdir(pathToDir)
  except OSError:
    logger.info("Deletion of the directory %s failed" % pathToDir)
  else:
    logger.info("Successfully deleted the directory %s" % pathToDir)

def list_files_in_dir(pathToDir:str, filterOptions: dict=dict()) -> List[str]:
  """Lists all files in targeted directory

  Args:
      pathToDir (str): path to targeted directory 
      filterOptions (dict, optional): Filter options for targeted files to search in targeted directory. Defaults to empty dict.

  Returns:
      List[str]: List of files found in the targeted directory
  """  
  files = []
  for each in [f for f in listdir(pathToDir) if isfile(join(pathToDir, f))]:
    if ('name_reg' in filterOptions) and \
      not bool(re.match(filterOptions['name_reg'], each)):
      continue
    if ('file_format' in filterOptions) and \
      (not filterOptions['file_format'].lower() in each.lower()):
      continue
    if ('name_like' in filterOptions) and \
      (not filterOptions['name_like'].lower() in each.lower()):
      continue
    files.append(each)
  return files

def unzip(inputFilenames:List[str], inputPath:str, outputPath:str) -> List[str]:
  """Unzip a zipfile

  Args:
      inputFilenames (List[str]): name of the zipped file
      inputPath (str): path to zipped file
      outputPath (str): path to place the unzipped files to

  Returns:
      List[str]: list of unzipped filenames
  """  
  outputFilenames = []
  for each in inputFilenames:
    zipfile = each if inputPath in each else f'{inputPath}/{each}' 
    with gzip.open(zipfile, 'rb') as fIn:
      filename, file_extension = os.path.splitext(each)
      filepath, filename = os.path.split(filename)
      unzipfile = filename if outputPath in filename else f'{outputPath}/{filename}'
      with open(unzipfile, 'wb') as fOut:
        shutil.copyfileobj(fIn, fOut)
        outputFilenames.append(filename)
  return outputFilenames
  


#!/usr/bin/env python3.8
# -*- coding: utf-8 -*-

"""This script contains functions to handle cloud storage
"""

from google.cloud import storage

def upload_to_bucket(remoteFilepath, localFilepath, bucketName:str):
    """ Upload data to a bucket"""
    storage_client = storage.Client()

    # get bucket to upload
    bucket = storage_client.get_bucket(bucketName)
    blob = bucket.blob(remoteFilepath)
    blob.upload_from_filename(localFilepath)
    print(blob)
    #returns a public url
    return blob.public_url



#!/usr/bin/env python3.8
# -*- coding: utf-8 -*-

"""This script contains helper functions to post message to slack channel
"""
from fp_th_di.logger import logger
import requests

class SlackBotMessenger:
  def post_json_to_slack(self, json:dict, channelUrl:str):
    """
      Sends data to Slack channel
      --------
      Args:
        filepath: string, required
          path to file to be sent to the targeted Slack channel
        channelUrl: string, required
          Webhook url of the targeted Slack channel.
    """

    try:
      requests.post(url=channelUrl, json=json)
    except Exception as e:
      logger.error(e)

  def post_message_to_slack(self, message:str, channelUrl:str):
    """
      Sends text message to Slack channel by webhook url
      Uses <@slack_member_id> to mention slack users
      --------
      Args:
        message: string, required
          message to send
        channelUrl: string, required
          Webhook url of the targeted Slack channel.
    """
    self.post_json_to_slack({'text': message}, channelUrl)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""This script contains function required for sending email
"""
# emailutils
import os
import re
import smtplib
import mimetypes
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email import encoders

def is_valid_email_address(email):
    ''' This function checks the format of email using regular expression
        PARAMETER:
            email (string): an email address
        RETURN:
            (re match object): an object for string that match the email pattern
    '''
    regex = '[^@]+@[^@]+\.[^@]+'
    return re.search(regex, email)


def send_mail(email_from, email_password, email_to, subject, content, is_html=False, attachments=None):
    ''' This function sends email with attachment (if any)
        PARAMETER:
            email_from (string): an email address for from part
            email_password (string): a password for email address, it should be app password generated in google account 
            email_to (string): an email add to send email to
            subject (string): a string as subject for email
            content (string): a string (html format is ok) as email content
            is_html (bool): True if the email content is in html format. Otherwise, False. Default value is False.
            attachments (list): a list of path to file to be attached
        RETURN:
            None
    '''
    
    # validate input
    if not bool(email_from):
        raise Exception('Invalid input - email_from is required')
    if not is_valid_email_address(email_from):
        raise Exception('Invalid input - email_from is invalid')
    if not bool(email_password):
        raise Exception('Invalid input - email_password is required')
    if not bool(email_to):
        raise Exception('Invalid input - email_to is required')
    
    # if email_to contains more than 1 email, split using ,
    if isinstance(email_to, str):
        email_to = email_to.split(',')
    else:
        email_to = ','.join(email_to).split(',')

    email_to = list(dict.fromkeys(email_to))
    cc_to = None
    if len(email_to) > 1:
        cc_to = ','.join(email_to[1:])
        email_to = email_to[0:1]
    email_to = ','.join(email_to)
    
    msg = MIMEMultipart()
    msg["From"] = email_from
    msg["To"] = email_to
    if cc_to is not None:
        msg["CC"] = cc_to

    msg["Subject"] = subject
    msg.attach(MIMEText(content, 'html' if is_html else 'plain')) 

    # process attachment
    if attachments:
        
        # for each file
        for fileToSend in attachments:
            
            # create attachment object associated to mimetype
            ctype, encoding = mimetypes.guess_type(fileToSend)
            if ctype is None or encoding is not None:
                ctype = "application/octet-stream"

            maintype, subtype = ctype.split("/", 1)
            if maintype == "text":
                fp = open(fileToSend)
                # Note: we should handle calculating the charset
                attachment = MIMEText(fp.read(), _subtype=subtype)
                fp.close()
            elif maintype == "image":
                fp = open(fileToSend, "rb")
                attachment = MIMEImage(fp.read(), _subtype=subtype)
                fp.close()
            elif maintype == "audio":
                fp = open(fileToSend, "rb")
                attachment = MIMEAudio(fp.read(), _subtype=subtype)
                fp.close()
            else:
                fp = open(fileToSend, "rb")
                attachment = MIMEBase(maintype, subtype)
                attachment.set_payload(fp.read())
                fp.close()
                encoders.encode_base64(attachment)

            filepath, filename = os.path.split(fileToSend)
            attachment.add_header("Content-Disposition", "attachment", filename=filename)
            msg.attach(attachment)

    # creates SMTP session
    s = smtplib.SMTP('smtp.gmail.com', 587)
    s.starttls()
    s.login(email_from, email_password)
    text = msg.as_string()
    
    # send the mail
    s.sendmail(email_from, email_to if cc_to is None else [email_to] + cc_to.split(','), text) 
    s.quit()



from fp_th_di.mail_box.email_attachment import EmailAttachment
from typing import List, Union
from fp_th_di.logger import logger
import datetime

class Email:
  def __init__(self, emailFrom, emailDate, emailSubject:str, 
  emailContent:str = None, emailAttachments: List[EmailAttachment] = []):
    self.emailFrom = emailFrom
    self.emailDate = emailDate # received datetime
    self.emailSubject = emailSubject
    self.emailContent = emailContent
    self.emailAttachments = emailAttachments

  def __str__(self) -> str:
    return str({
        'emailFrom': self.emailFrom,
        'emailDate': self.get_received_on_datetime(),
        'emailSubject': self.emailSubject,
        'emailContent': self.emailContent,
        'emailAttachments': [str(each_attachment) \
          for each_attachment in self.emailAttachments] 
    })

  def add_attachment(self, attachment:EmailAttachment) -> None:
    """add an attachment to attachment list 

    Args:
        attachment (EmailAttachment)
    """    
    if attachment not in self.emailAttachments:
      self.emailAttachments.append(attachment)
    
  def has_attachment(self) -> bool:
    """Check whether email has attachment

    Returns:
        bool: True if it has at least one attachments, otherwise false
    """    
    return bool(self.emailAttachments)

  def get_attachment(self, index:int = 0) -> Union[EmailAttachment, None]:
    """Get email attachment from attachment list

    Args:
        index (int, optional): index of attachment. Defaults to 0.

    Returns:
        Union[EmailAttachment, None]: An instance of EmailAttachment if exists; otherwise, None
    """    
    if self.has_attachment():
      return self.emailAttachments[index]
    return None

  def get_received_on_datetime(self) -> datetime.datetime:
    """Get datetime timestamp of emailDate

    Returns:
        datetime.datetime: datetime of emailDate 
    """    
    return (datetime.datetime(self.emailDate[0], 
      self.emailDate[1], self.emailDate[2], self.emailDate[3], 
      self.emailDate[4], self.emailDate[5]) + datetime.timedelta(hours=0))

  def was_received_on(self, tergetedDate:datetime.datetime) -> bool:
    """Check whether the email was received on a targeted date or not

    Args:
        tergetedDate (datetime.datetime): targeted date to compare

    Returns:
        bool: True if email's received on targeted date; otherwise, False
    """    
    logger.info(self.get_received_on_datetime())
    return self.get_received_on_datetime().date == tergetedDate.date



#!/usr/bin/env python3.8
# -*- coding: utf-8 -*-

"""This script contains functions to handle secret values
"""

from google.cloud import secretmanager
import hashlib


def secret_hash(secretValue: str):
  """Get the sha224 hash of the secret value

    Args:
      secretValue (str): secret string to hash

    Returns:
        _Hash: sha224 of utf-8 secret value
    """
  return hashlib.sha224(bytes(secretValue, "utf-8")).hexdigest()

def get_secret_value_from_google(projectId: str, secretId: str, versionId='latest') -> str:
  """Get secret from Google Secret Manager

  Args:
      projectId (str): project id or name
      secretId (str): secret id or name
      versionId (str, optional): secret version. Defaults to 'latest'.

  Returns:
      str: secret stored in Google Secret Manager
  """
  client = secretmanager.SecretManagerServiceClient()

  # resource name of the targeted version
  name = f'projects/{projectId}/secrets/{secretId}/versions/{versionId}'

  # access the version
  response = client.access_secret_version(name=name)

  # decode payload
  return response.payload.data.decode('UTF-8')
